"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const chai_1 = require("chai");
const Uuid = require("uuid");
const etcd = require("../lib/index");
const rx = require("rxjs");
/*
function param(arr: string[], uuid: string): string[] {
  return arr.concat(['--etcd-cluster-id', uuid, '--etcd-url', 'http://localhost:2379']);
}
*/
describe('etcd', function () {
    this.timeout(2000);
    before(() => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://localhost:2379'
        ]);
        console.log('etcd Cluster Booted...0');
        let etc = etcd.EtcdPromise.create(wc);
        console.log('etcd Cluster Booted...1');
        yield etc.connect();
        console.log('etcd Cluster Booted...2');
        return Promise.resolve('done');
    }));
    it('AsyncPromise Blocking', () => __awaiter(this, void 0, void 0, function* () {
        let value = 1;
        let out = new Promise((r, j) => __awaiter(this, void 0, void 0, function* () {
            yield new Promise((rr, jj) => { setTimeout(rr, 10); });
            ++value;
        }));
        out = null; // WTF Lint
        chai_1.assert.equal(value, 1);
        yield new Promise((rr, jj) => { setTimeout(rr, 15); });
        chai_1.assert.equal(value, 2);
        return Promise.resolve('done');
    }));
    it('selfStat one good', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://localhost:2379'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.clusterState();
        chai_1.assert.equal(1, ret.length);
        // console.log(ret[0])
        chai_1.assert.equal(ret[0].isOk(), true);
        chai_1.assert.equal(ret[0].url, 'http://localhost:2379');
        return Promise.resolve('done');
    }));
    it('selfStat two same good', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://localhost:2379',
            '--etcd-url', 'http://localhost:2379'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.clusterState();
        chai_1.assert.equal(2, ret.length);
        chai_1.assert.equal(ret[0].isOk(), true);
        chai_1.assert.equal(ret[0].url, 'http://localhost:2379');
        chai_1.assert.equal(ret[1].isOk(), true);
        chai_1.assert.equal(ret[1].url, 'http://localhost:2379');
        return Promise.resolve('done');
    }));
    it('selfStat same good', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://localhost:2379',
            '--etcd-url', 'http://localhost:2379',
            '--etcd-url', 'http://localhost:2379',
            '--etcd-url', 'http://localhost:2379'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.clusterState();
        chai_1.assert.equal(4, ret.length);
        chai_1.assert.equal(ret[0].isOk(), true);
        chai_1.assert.equal(ret[0].url, 'http://localhost:2379');
        return Promise.resolve('done');
    }));
    it('selfStat connect', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://localhost:2379'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.connect();
        chai_1.assert.equal(ret.isOk(), true, 'isok');
        chai_1.assert.equal(ret.url, 'http://localhost:2379', 'url');
        return Promise.resolve('done');
    }));
    it('selfStat clusterState bad', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://169.254.99.94'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.clusterState();
        chai_1.assert.equal(1, ret.length);
        chai_1.assert.equal(ret[0].isOk(), false);
        return Promise.resolve('done');
    }));
    it('selfStat connect bad', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://169.254.99.94',
            '--etcd-url', 'http://169.254.99.95'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.connect();
        chai_1.assert.equal(ret.isOk(), false);
        return Promise.resolve('done');
    }));
    it('selfStat connect good', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://localhost:2379'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.connect();
        chai_1.assert.equal(ret.isOk(), true);
        return Promise.resolve('done');
    }));
    it('selfStat connect two good', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://localhost:2379',
            '--etcd-url', 'http://127.0.0.1:2379'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.connect();
        chai_1.assert.equal(ret.isOk(), true);
        return Promise.resolve('done');
    }));
    it('selfStat connect two good', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://localhost:2379',
            '--etcd-url', 'http://169.254.55.99',
            '--etcd-url', 'http://127.0.0.1:2379'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.connect();
        chai_1.assert.equal(ret.isOk(), true);
        return Promise.resolve('done');
    }));
    it('selfStat one good one bad', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://localhost:2379',
            '--etcd-url', 'http://169.254.99.94'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.clusterState();
        chai_1.assert.equal(2, ret.length);
        chai_1.assert.equal(ret[0].isOk(), true);
        chai_1.assert.equal(ret[1].isOk(), false);
        return Promise.resolve('done');
    }));
    it('selfStat two good one bad', () => __awaiter(this, void 0, void 0, function* () {
        let wc = etcd.Config.start([
            '--etcd-req-timeout', '50',
            '--etcd-url', 'http://localhost:2379',
            '--etcd-url', 'http://169.254.99.94',
            '--etcd-url', 'http://127.0.0.1:2379'
        ]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.clusterState();
        chai_1.assert.equal(3, ret.length);
        chai_1.assert.equal(ret[0].isOk(), true);
        chai_1.assert.equal(ret[1].isOk(), false);
        chai_1.assert.equal(ret[2].isOk(), true);
        return Promise.resolve('done');
    }));
    it('empty-list', () => __awaiter(this, void 0, void 0, function* () {
        let uuid = Uuid.v4().toString();
        let wc = etcd.Config.start(['--etcd-cluster-id', uuid]);
        let etc = etcd.EtcdPromise.create(wc);
        let lst = yield etc.list('');
        chai_1.assert.isTrue(lst.isErr());
        yield etc.mkdir('');
        lst = yield etc.list('');
        chai_1.assert.isFalse(lst.isErr());
        yield etc.mkdir('meno');
        lst = yield etc.list('');
        chai_1.assert.isTrue(lst.value[0].key.endsWith('meno'));
    }));
    it('mkdir', () => __awaiter(this, void 0, void 0, function* () {
        let uuid = Uuid.v4().toString();
        let wc = etcd.Config.start(['--etcd-cluster-id', uuid]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.mkdir('meno');
        chai_1.assert.equal(ret.isErr(), false);
        ret = yield etc.mkdir('meno');
        chai_1.assert.equal(ret.isErr(), true);
        let lst = yield etc.list('meno');
        chai_1.assert.equal(lst.isErr(), false);
        chai_1.assert.equal(lst.value.length, 0);
        ret = yield etc.rmdir('meno');
        chai_1.assert.equal(ret.isOk(), true);
    }));
    it('rmdir', () => __awaiter(this, void 0, void 0, function* () {
        let uuid = Uuid.v4().toString();
        let wc = etcd.Config.start(['--etcd-cluster-id', uuid]);
        let etc = etcd.EtcdPromise.create(wc);
        yield etc.mkdir('meno');
        yield etc.mkdir('meno/geheim');
        yield etc.mkdir('meno/geheim/ganz');
        let ret = yield etc.rmdir('meno', { recursive: true });
        chai_1.assert.equal(ret.isOk(), true);
        yield etc.mkdir('meno');
        let lst = yield etc.list('meno');
        chai_1.assert.equal(lst.isErr(), false);
        chai_1.assert.equal(lst.value.length, 0);
        ret = yield etc.rmdir('meno', { recursive: true });
        chai_1.assert.equal(ret.isOk(), true);
    }));
    it('get/setRaw', () => __awaiter(this, void 0, void 0, function* () {
        let uuid = Uuid.v4().toString();
        let wc = etcd.Config.start(['--etcd-cluster-id', uuid]);
        let etc = etcd.EtcdPromise.create(wc);
        let ret = yield etc.getRaw('hammer/murk');
        chai_1.assert.equal(ret.isErr(), true);
        ret = yield etc.setRaw('hammer/murk', 'Hello World');
        chai_1.assert.equal(ret.isErr(), false);
        ret = yield etc.getRaw('hammer/murk');
        chai_1.assert.equal(ret.isOk(), true);
        // console.log(ret.node)
        chai_1.assert.equal(ret.node.value, 'Hello World');
    }));
    function leftPad(n, pad) {
        return ('' + (Math.pow(10, pad) + n)).substr(1);
    }
    function changeWriterThen(source, ccw, cnt, done) {
        let nested = 0;
        ccw.subscribe((er) => {
            nested++;
            let node = er.node;
            if (er.node.dir) {
                node = er.node.nodes.sort((a, b) => {
                    if (a.key < b.key) {
                        return -1;
                    }
                    if (a.key > b.key) {
                        return 1;
                    }
                    return 0;
                })[er.node.nodes.length - 1];
            }
            // console.log(cnt.cnt, node.key, node.value, nested);
            chai_1.assert.isTrue(node.key.endsWith(`wait-for-change/hallo-${leftPad(cnt.cnt, 5)}`), 'not match');
            chai_1.assert.equal(`vallo-${leftPad(cnt.cnt, 5)}`, node.value, 'funny value');
            cnt.cnt++;
            if (cnt.cnt && (cnt.cnt % 4) == 0) {
                done();
            }
            else {
                updateData(source, ccw, cnt, done, false);
            }
            --nested;
        }, () => {
            chai_1.assert.fail('should never called');
        });
    }
    function updateData(source, ccw, cnt, done, install) {
        setTimeout(() => {
            source.setRaw(`wait-for-change/hallo-${leftPad(cnt.cnt, 5)}`, `vallo-${leftPad(cnt.cnt, 5)}`).then(() => {
                if (install) {
                    changeWriterThen(source, ccw, cnt, done);
                }
            }).catch(() => {
                chai_1.assert.fail('should never called');
            });
        }, (cnt.cnt % 1000) * 90);
    }
    it('wait-for-change', (done) => {
        const uuid = Uuid.v4().toString();
        const wc = etcd.Config.start([
            '--etcd-cluster-id', uuid,
            '--etcd-req-timeout', '200'
        ]);
        const source = etcd.EtcdPromise.create(wc);
        source.mkdir('wait-for-change').then(() => {
            const ccw = source.createChangeWaiter('wait-for-change', { recursive: true });
            const cnt = { cnt: 0 };
            updateData(source, ccw, cnt, () => {
                ccw.cancel();
                cnt.cnt = 1000;
                setTimeout(() => updateData(source, ccw, cnt, () => {
                    ccw.cancel();
                    setTimeout(() => {
                        source.setRaw(`wait-for-change/cancelled`, `stopped`).then(() => {
                            done();
                        }).catch(() => {
                            chai_1.assert.fail('never called');
                        });
                    }, 100);
                }, true), 300);
            }, true);
        });
    });
    function upsetTester(obEtcd, done, apply, test) {
        const lifeCycle = {
            apply: 0,
            next: 0,
            error: 0,
            complete: 0
        };
        const upset = etcd.Upset.create(obEtcd).upSet('upset', (inp, obs) => {
            // console.log('apply');
            apply(lifeCycle, inp, obs);
        });
        const completed = new rx.Subject();
        upset.subscribe(() => {
            // console.log('next');
            lifeCycle.next++;
        }, (error) => {
            // console.log('error');
            lifeCycle.error++;
        }, () => {
            // console.log('complete');
            lifeCycle.complete++;
            completed.complete();
        });
        completed.subscribe(null, null, () => {
            // console.log('2-next', lifeCycle);
            test(lifeCycle, done);
        });
    }
    it('upset-empty-noopt', (done) => {
        const uuid = Uuid.v4().toString();
        const wc = etcd.Config.start([
            '--etcd-cluster-id', uuid,
            '--etcd-req-timeout', '200'
        ]);
        const obEtcd = etcd.EtcdObservable.create(wc);
        upsetTester(obEtcd, done, (lifeCycle, inp, outer) => {
            chai_1.assert.isNull(inp);
            lifeCycle.apply++;
            outer.complete();
        }, (lifeCycle, _done) => {
            chai_1.assert.equal(lifeCycle.apply, 1);
            chai_1.assert.equal(lifeCycle.next, 0);
            chai_1.assert.equal(lifeCycle.error, 0);
            chai_1.assert.equal(lifeCycle.complete, 1);
            obEtcd.getRaw('upset').subscribe((er) => {
                if (er.isErr() && er.err.etcErr.errorCode == 100) {
                    return;
                }
                chai_1.assert.fail('never reached');
            }, (err) => {
                chai_1.assert.fail('no error expected');
            }, () => {
                _done();
            });
        });
    });
    it('upset-from-empty-set', (done) => {
        const uuid = Uuid.v4().toString();
        const wc = etcd.Config.start([
            '--etcd-cluster-id', uuid,
            '--etcd-req-timeout', '200'
        ]);
        const obEtcd = etcd.EtcdObservable.create(wc);
        upsetTester(obEtcd, done, (lifeCycle, inp, outer) => {
            chai_1.assert.isNull(inp);
            lifeCycle.apply++;
            outer.next({ 'hello': 'world' });
            outer.complete();
        }, (lifeCycle, _done) => {
            chai_1.assert.equal(lifeCycle.apply, 1);
            chai_1.assert.equal(lifeCycle.next, 1);
            chai_1.assert.equal(lifeCycle.error, 0);
            chai_1.assert.equal(lifeCycle.complete, 1);
            // wc.log.info('empty-set:0');
            obEtcd.getRaw('upset').subscribe((er) => {
                // wc.log.info('empty-set:1');
                if (er.isErr()) {
                    // wc.log.info('empty-set:2');
                    chai_1.assert.fail('not found');
                }
                // wc.log.info('empty-set:3');
                chai_1.assert.deepEqual(JSON.parse(er.node.value), { 'hello': 'world' });
            }, (err) => {
                // wc.log.info('empty-set:4');
                chai_1.assert.fail('no error expected');
            }, () => {
                // wc.log.info('empty-set:5');
                _done();
            });
        });
    });
    it('upset-from-nonempty-set', (done) => {
        const uuid = Uuid.v4().toString();
        const wc = etcd.Config.start([
            '--etcd-cluster-id', uuid,
            '--etcd-req-timeout', '200'
        ]);
        const obEtcd = etcd.EtcdObservable.create(wc);
        upsetTester(obEtcd, done, (lifeCycle, inp, outer) => {
            chai_1.assert.isNull(inp);
            lifeCycle.apply++;
            outer.next({ 'hello': 'world' });
            outer.complete();
        }, (lifeCycle, _done) => {
            chai_1.assert.equal(lifeCycle.apply, 1);
            chai_1.assert.equal(lifeCycle.next, 1);
            chai_1.assert.equal(lifeCycle.error, 0);
            chai_1.assert.equal(lifeCycle.complete, 1);
            obEtcd.getRaw('upset').subscribe((er) => {
                if (er.isErr()) {
                    chai_1.assert.fail('not found');
                }
                chai_1.assert.deepEqual(JSON.parse(er.node.value), { 'hello': 'world' });
            }, (err) => {
                chai_1.assert.fail('no error expected');
            }, () => {
                upsetTester(obEtcd, done, (_lifeCycle, inp, outer) => {
                    chai_1.assert.deepEqual(inp, { 'hello': 'world' });
                    _lifeCycle.apply++;
                    outer.next({ 'world': 'hello' });
                    outer.complete();
                }, (_lifeCycle, __done) => {
                    chai_1.assert.equal(_lifeCycle.apply, 1);
                    chai_1.assert.equal(_lifeCycle.next, 1);
                    chai_1.assert.equal(_lifeCycle.error, 0);
                    chai_1.assert.equal(_lifeCycle.complete, 1);
                    obEtcd.getRaw('upset').subscribe((er) => {
                        if (er.isErr()) {
                            chai_1.assert.fail('not found');
                        }
                        chai_1.assert.deepEqual(JSON.parse(er.node.value), { 'world': 'hello' });
                    }, (err) => {
                        chai_1.assert.fail('no error expected');
                    }, () => {
                        __done();
                    });
                });
            });
        });
    });
    it('upset-set-concurrent', (done) => {
        const uuid = Uuid.v4().toString();
        const wc = etcd.Config.start([
            '--etcd-cluster-id', uuid,
            '--etcd-req-timeout', '200'
        ]);
        const obEtcd = etcd.EtcdObservable.create(wc);
        upsetTester(obEtcd, done, (lifeCycle, inp, outer) => {
            if (lifeCycle.apply == 0) {
                // console.log('apply:0');
                upsetTester(obEtcd, done, (_lifeCycle, _inp, _outer) => {
                    chai_1.assert.isNull(_inp);
                    _lifeCycle.apply++;
                    _outer.next({ 'hello': 'world' });
                    _outer.complete();
                }, (_lifeCycle, __done) => {
                    chai_1.assert.equal(_lifeCycle.apply, 1);
                    chai_1.assert.equal(_lifeCycle.next, 1);
                    chai_1.assert.equal(_lifeCycle.error, 0);
                    chai_1.assert.equal(_lifeCycle.complete, 1);
                    obEtcd.getRaw('upset').subscribe((er) => {
                        if (er.isErr()) {
                            chai_1.assert.fail('not found');
                        }
                        chai_1.assert.deepEqual(JSON.parse(er.node.value), { 'hello': 'world' });
                        // wc.log.info('inner.upset:next');
                    }, (err) => {
                        chai_1.assert.fail('no error expected');
                    }, () => {
                        // wc.log.info('inner.upset:complete:0');
                        chai_1.assert.isNull(inp);
                        lifeCycle.apply++;
                        outer.next({ 'world': 'hello' });
                        outer.complete();
                        // wc.log.info('inner.upset:complete:1');
                    });
                });
            }
            else {
                // console.log('apply:1');
                chai_1.assert.deepEqual(inp, { 'hello': 'world' });
                lifeCycle.apply++;
                outer.next({ 'world': 'hello' });
                outer.complete();
            }
        }, (lifeCycle, _done) => {
            chai_1.assert.equal(lifeCycle.apply, 2);
            chai_1.assert.equal(lifeCycle.next, 1);
            chai_1.assert.equal(lifeCycle.error, 0);
            chai_1.assert.equal(lifeCycle.complete, 1);
            obEtcd.getRaw('upset').subscribe((er) => {
                if (er.isErr()) {
                    chai_1.assert.fail('not found');
                }
                chai_1.assert.deepEqual(JSON.parse(er.node.value), { 'world': 'hello' });
            }, (err) => {
                chai_1.assert.fail('no error expected');
            }, () => {
                _done();
            });
        });
    });
});
//# sourceMappingURL=etcd-test.js.map